import pexpect
import subprocess
import ujson
from collections import defaultdict

from edge import eventlog

_HASP_FEATURE_HAS_EXPIRED_TEXT = "Feature has expired"
_HASP_FEATURE_NOT_FOUND_TEXT = "Feature not found"
_HASP_FEATURE_NOT_FOUND_CODE = "H0031"
_HASP_KEY_NOT_FOUND_TEXT = "Sentinel key not found"
_HASP_KEY_NOT_FOUND_CODE = "H0007"


def lsvpwc(by_group):
    """
    :type by_group: bool
    :rtype: list of dict
    """
    command_line = '/opt/edge/bin/lsvpwc -m /opt/edge/libexec/lpr/aorp/modules/ -t json'
    if by_group:
        command_line += ' -g'

    lsvpwc_proc = pexpect.spawn(command_line)
    try:
        lsvpwc_proc.expect(pexpect.EOF)
        output = lsvpwc_proc.before
        if not output:
            raise RuntimeError('fail: lsvpwc (stdout is empty)')

    finally:
        lsvpwc_proc.close()

    try:
        vpw_license_group_jnode_list = ujson.loads(output)

    except ValueError:
        vpw_license_group_jnode_list = []
        if _HASP_FEATURE_HAS_EXPIRED_TEXT in output:
            eventlog.log_d(_HASP_FEATURE_HAS_EXPIRED_TEXT)
            return vpw_license_group_jnode_list

        elif _HASP_FEATURE_NOT_FOUND_TEXT in output or _HASP_FEATURE_NOT_FOUND_CODE in output:
            eventlog.log_d(_HASP_FEATURE_NOT_FOUND_TEXT)
            return vpw_license_group_jnode_list

        elif _HASP_KEY_NOT_FOUND_TEXT in output or _HASP_KEY_NOT_FOUND_CODE in output:
            eventlog.log_d(_HASP_KEY_NOT_FOUND_TEXT)
            return vpw_license_group_jnode_list

        else:
            raise RuntimeError('fail: loads (%s)' % output)

    return vpw_license_group_jnode_list


def get_ui_countries_jnode(vpw_license_countries):
    """
    :type vpw_license_countries: str
    :rtype: dict
    """
    # if lsvpwc returns '' (empty templates), license has expired
    if not vpw_license_countries:
        return []

    # if lsvpwc returns * activate all world
    if '*' in vpw_license_countries:
        vpw_license_countries = ' '

    stdout, stderr = subprocess.Popen(
        "/opt/edge/bin/lsvpwi -m /opt/edge/libexec/lpr/aorp/modules/ -t cut -s id %s" % vpw_license_countries,
        stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True).communicate()

    if stderr:
        eventlog.log_d('fail: lsvpwi (%s)' % stderr)
        return None

    raw_template_set_as_str = stdout
    raw_template_set_as_list = raw_template_set_as_str.split('\n')[:-1]

    template_set = []
    for raw_template in raw_template_set_as_list:
        raw_items = raw_template.split('.')
        raw_item_count = len(raw_items)
        if raw_item_count == 2:
            final_items = [raw_items[0], "0", raw_items[1]]
            template_set.append('-'.join(final_items))

        elif raw_item_count == 3:
            final_items = raw_items
            template_set.append('-'.join(final_items))

        else:
            raise NotImplementedError('fail: %s' % raw_template_set_as_list)

    country_to_region_to_template_map = defaultdict(lambda: defaultdict(list))
    template_set = template_set[:-1]

    for template in template_set:
        country, region, _ = template.split('-')
        country_to_region_to_template_map[country][region].append(template)

    child = pexpect.spawn("/opt/edge/bin/lpr-iso3166", template_set)
    try:
        child.expect(pexpect.EOF)
        stdout = child.before
        if not stdout:
            raise RuntimeError('fail: lpr-iso3166 (stdout is empty)')

    finally:
        child.close()

    country_region_to_titles_map = ujson.loads(stdout)

    countries_jnode = []
    for country, region_to_templates_map in country_to_region_to_template_map.iteritems():
        country_title = None
        regions_jnode = []
        for region, templates in region_to_templates_map.iteritems():
            region_jnode = {
                "templates": templates
            }

            country_title = country_region_to_titles_map[country][region]['country-title']
            region_title = country_region_to_titles_map[country][region].get('region-title')
            if region_title:
                region_jnode['title'] = region_title

            regions_jnode.append(region_jnode)

        if 'title' in regions_jnode[0]:
            regions_jnode = sorted(regions_jnode, key=lambda region_item: region_item['title'])

        country_jnode = {
            'title': country_title,
            'regions': regions_jnode
        }

        countries_jnode.append(country_jnode)

    countries_jnode.sort(key=lambda country_item: country_item['title'])

    return countries_jnode
