#!/usr/bin/python
import ujson
import sys

import edge.admin.ui_types as ui_types
from edge import eventlog
from edge.admin.reply_helpers import *
import edge.pipeline_helpers as pipeline_helpers

import pipeline.mv_pipeline_creator as mv_pipeline_creator
import pipeline.rtsp_pipeline_creator as rtsp_pipeline_creator
import pipeline.ui_pipeline_set as ui_pipeline_set
from license.lpr._helpers import get_used_license_count, get_total_license_count
from pipeline.ui_pipeline import UiPipeline, get_pipeline_ext_jnode
import pipeline._camera_discover as camera_discover


class PipelineError(RuntimeError):
    pass


def handle_define_videosource():
    return reply_done(camera_discover.discover_cameras())


def _get_new_pid(pipeline_jnode_set):
    """
    :type pipeline_jnode_set: list of dict
    :rtype: str | None
    """
    used_ui_pid_set = set(map(lambda pipeline_jnode: int(pipeline_jnode['id']), pipeline_jnode_set))

    ui_pid = None
    for ui_pid_candidate in xrange(0, 32):
        if ui_pid_candidate not in used_ui_pid_set:
            ui_pid = str(ui_pid_candidate)
            eventlog.log_d("new-pipeline-id: %s" % ui_pid)
            break
    return ui_pid


def _run(query_args):
    """
    :type query_args: edge.admin.ui_types.UiQueryArgs
    :raises: PipelineError
    """
    # handle ../pipeline
    if query_args.query_string:
        eventlog.log_d("fail: bad-request")
        return reply_fail(REPLY_BAD_REQUEST_400)

    if query_args.http_method == 'GET':

        if len(query_args.path_info) != 2:
            eventlog.log_d("fail: not-found (%s)" % query_args.path_info)
            return reply_fail(REPLY_NOT_FOUND_404)

        if query_args.path_info[1] == 'cameras':
            return handle_define_videosource()
        elif query_args.path_info[1] != 'info':
            return reply_fail(REPLY_NOT_FOUND_404)

        if query_args.raw_body:
            eventlog.log_d("fail: bad-request")
            return reply_fail(REPLY_BAD_REQUEST_400)

        pipeline_set_jnode = pipeline_helpers.get_pipeline_jnode_set2()
        if pipeline_set_jnode is None:
            eventlog.log_d("fail: internal-server-error (invalid pipeline-set-jnode)")
            raise PipelineError

        reply_jnode = get_pipeline_ext_jnode(pipeline_set_jnode)

        return reply_done(reply_jnode)

    elif query_args.http_method == 'POST':
        if len(query_args.path_info) != 2:
            eventlog.log_d("fail: not-found (%s)" % query_args.path_info)
            return reply_fail(REPLY_NOT_FOUND_404)

        if query_args.path_info[1] != 'units':
            return reply_fail(REPLY_NOT_FOUND_404)

        # noinspection PyBroadException
        try:
            total_thread_count = get_total_license_count()
            used_thread_count = get_used_license_count()
            eventlog.log_d('license total_thread_count: %s, used_thread_count: %s' %
                               (total_thread_count, used_thread_count))

        except RuntimeError:
            eventlog.log_d('fail: get license')
            raise PipelineError

        if used_thread_count >= total_thread_count:
            eventlog.log_d('fail: free lpr licence is not available')
            return reply_fail2(REPLY_SERVICE_UNAVAILABLE_503, error_code='lpr-licence-absent',
                               error_text='free lpr licence is not available')

        else:
            eventlog.log_d('License OK')

        # define ui_pid of new channel
        pipeline_jnode_set = pipeline_helpers.get_pipeline_jnode_set2()
        if pipeline_jnode_set is None:
            eventlog.log_d("fail: internal-server-error")
            raise PipelineError

        ui_pid = _get_new_pid(pipeline_jnode_set)
        if ui_pid is None:
            eventlog.log_d("fail: service-unavailable")
            return reply_fail(REPLY_SERVICE_UNAVAILABLE_503)

        if query_args.raw_body:
            reply_jnode = mv_pipeline_creator.create_mv_pipeline(query_args.raw_body, ui_pid)
            return reply_jnode

        else:
            reply_jnode = rtsp_pipeline_creator.create_rtsp_pipeline(ui_pid)
            return reply_jnode

    else:
        eventlog.log_d("fail: method-not-allowed")
        return reply_fail(REPLY_METHOD_NOT_ALLOWED_405)


# noinspection PyBroadException
def _main():
    eventlog.open_log('cgi-controller', 'debug')
    exit_code = 1
    try:
        reply_jnode = reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)
        try:
            query_args = ui_types.UiQueryArgs.deserialize('/')
            reply_jnode = _run(query_args)
            exit_code = 0
        except PipelineError:
            pass

        except:
            eventlog.log_x()

        if reply_jnode.get('envelope', True):
            if 'envelope' in reply_jnode:
                del reply_jnode['envelope']
            print ujson.dumps(reply_jnode)

        else:
            if reply_jnode.get('status') == REPLY_SUCCESS_200:
                print ujson.dumps(reply_jnode['result'])

        sys.stderr.write(str(reply_jnode['status']))

        if exit_code != 0:
            eventlog.log_d('fail: internal-server-error')

    except:
        eventlog.log_x()

    finally:
        eventlog.close_log()

    sys.exit(exit_code)


if __name__ == '__main__':
    _main()
