import os
import ujson

from edge import path_helpers as ph, pipeline_helpers as oh
from edge.com import fs, net
from edge.admin.ui_types import BaseCreator


def _reconfigure_es(endpoints):
    """
    :type endpoints: list[str]
    :rtype: bool
    """
    for _ in [1, 2, 3]:
        try:
            net.query_es_config(endpoints, tout_sec=1)
            return True
        except net.NetError:
            pass

    return False


def _mk_react_path(ui_pid):
    """
    :type ui_pid: str
    :rtype: str
    """
    return '/var/spool/edge/system/sc{0}-es65535-fact'.format(ui_pid)


def es_try_get_creator_ctor(ui_entry_point, used_ui_modules, used_edge_modules, avaliable_ui_modules):
    """
    :type ui_entry_point: ph.UiIdentity
    :type used_ui_modules: set of ph.UiIdentity
    :type used_edge_modules: set of ph.EdgeIdentity
    :type avaliable_ui_modules: set of ph.UiIdentity
    :rtype: None | (str, ph.UiIdentity, dict of ph.UiIdentity: dict) -> BaseCreator
    """
    creator_ctor = None
    # speed
    if ui_entry_point in [ph.Pipeline.Speed.Ui.PIPELINE]:
        used_edge_modules.update([ph.Pipeline.Speed.Edge.EVENTSTORE_65535])
        creator_ctor = EsCreator.create

    # mv
    elif ui_entry_point in [ph.Pipeline.Mv.Ui.PIPELINE]:
        used_edge_modules.update([ph.Pipeline.Mv.Edge.EVENTSTORE_65535])
        creator_ctor = EsCreator.create

    # rtsp
    elif ui_entry_point in [ph.Pipeline.Rtsp.Ui.PIPELINE]:
        used_edge_modules.update([ph.Pipeline.Rtsp.Edge.EVENTSTORE_65535])
        creator_ctor = EsCreator.create

    return creator_ctor


class EsCreator(BaseCreator):
    def __init__(self, id_main_setupstore):
        """
        :type id_main_setupstore: str, dict, dict | None, str
        """
        BaseCreator.__init__(self)
        self._id_main_setupstore = id_main_setupstore

    @staticmethod
    def create(ui_pid, ui_entry_point, ui_identity_to_jnode_map):
        """
        :type ui_pid: str
        :type ui_entry_point: ph.UiIdentity
        :type ui_identity_to_jnode_map: dict of ph.UiIdentity: dict
        :rtype: EsCreator
        """
        # speed
        if ui_entry_point in [ph.Pipeline.Rtsp.Ui.PIPELINE, ph.Pipeline.Mv.Ui.PIPELINE,
                              ph.Pipeline.Speed.Ui.PIPELINE]:
            id_main_setupstore = EsCreator._serialize(ui_pid)
            creator = EsCreator(id_main_setupstore)

        else:
            raise NotImplemented('pipeline not supported')

        return creator

    @staticmethod
    def _serialize(ui_pid):
        edge_id = '65535'

        origins_pid_set = oh.get_origin_pids_set()
        origins_pid_set.add(ui_pid)

        edge_jnode_dir = ph.get_edge_jnode_dirpath(ph.EDGE_TYPE_EVENTSTORE, edge_id)
        main_conf_abspath = os.path.join(edge_jnode_dir, 'main.conf')
        with open(main_conf_abspath, 'r') as json_file:
            old_main_conf_jnode = ujson.load(json_file)

        new_main_conf = {
            "api": oh.get_es_uds_endpoint(),
            "db": ph.DB_JSON_FILEPATH,
            "monitoring": {
                "stat-path": "/var/spool/edge/system/zabbix/es65535",
                "stat-tout": 60
            },
            "fact-react": map(_mk_react_path, origins_pid_set),
            "systest": old_main_conf_jnode.get('systest', False)
        }

        setupstore_conf_text = ph.get_setupstore_conf_text(edge_id)
        return edge_id, new_main_conf, setupstore_conf_text

    def serialize(self):
        edge_id, main_conf, setupstore_conf_text = self._id_main_setupstore
        edge_jnode_dir = ph.get_edge_jnode_dirpath(ph.EDGE_TYPE_EVENTSTORE, edge_id)

        if not os.path.exists(edge_jnode_dir):
            oh.silent_mkdir_r(edge_jnode_dir)

        conf_path = fs.join_path(edge_jnode_dir, 'main.conf')
        with open(conf_path, 'w') as conf:
            ujson.dump(main_conf, conf, indent=2)

        conf_path = fs.join_path(edge_jnode_dir, 'setupstore.conf')
        with open(conf_path, 'w') as conf:
            conf.write(setupstore_conf_text)

        if not _reconfigure_es(main_conf['fact-react']):
            oh.restart_eventstore()


class EsDeleter:
    def __init__(self, main_conf_filepath, ui_pid):
        """
        :type main_conf_filepath: str
        :type ui_pid: str
        """
        self._main_conf_filepath = main_conf_filepath
        self._ui_pid = ui_pid

    @staticmethod
    def create(ui_pid, ui_entry_point):
        """
        :type ui_pid: str
        :type ui_entry_point: ph.UiIdentity
        :rtype: EsDeleter | None
        """
        if ui_entry_point in [ph.Pipeline.Rtsp.Ui.PIPELINE, ph.Pipeline.Mv.Ui.PIPELINE,
                              ph.Pipeline.Speed.Ui.PIPELINE]:

            edge_jnode_dir = ph.get_edge_jnode_dirpath(ph.EDGE_TYPE_EVENTSTORE, '65535')
            main_conf_path = os.path.join(edge_jnode_dir, 'main.conf')
            deleter = EsDeleter(main_conf_path, ui_pid)
            return deleter

        else:
            raise NotImplemented

    def run(self):
        main_conf = fs.load_json2(self._main_conf_filepath)

        fact_react = filter(
            lambda endpoint: endpoint != _mk_react_path(self._ui_pid),
            main_conf['fact-react'])

        main_conf['fact-react'] = fact_react

        fs.write_json_file(main_conf, self._main_conf_filepath)

        if not _reconfigure_es(main_conf['fact-react']):
            oh.restart_eventstore()
