import jsonschema
import ujson

from edge import eventlog, path_helpers as ph, pipeline_helpers as helpers
from edge.admin.ui_types import LockError
from edge.admin import reply_helpers as rh, ui_types as ui
from edge.com import fs
from edge.types import FormatTemplate

import edge_creator as creator
from ui_pipeline_set import get_pipeline_type


def _validate_format(jnode):
    """
    :type jnode: list
    :rtype: bool
    """
    key_to_true_map = {
        'datetime': True,
        'datetime-ms': True,
        'post-name': True,
        'geolocation': True,
        'vendor': True,
        'model': True,
        'serial': True
    }

    for row in jnode:
        for col in row['cols']:
            try:
                FormatTemplate(template=col['format']).substitute(key_to_true_map)
            except KeyError as x:
                eventlog.log_d('fail: invalid key (%s)' % x.message)
                return False

    return True


def _handle_get(ui_pid, entry_point):
    """
    :type ui_pid: str
    :type entry_point: ph.UiIdentity
    :rtype: dict
    """
    videsource_status = helpers.get_videsource_status(ui_pid, entry_point.id)
    if videsource_status is None:
        eventlog.log_e('fail: get videosource status')
        return rh.reply_fail2(
            rh.REPLY_SERVICE_UNAVAILABLE_503,
            error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
            error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

    elif videsource_status != 0:
        eventlog.log_e('fail: videosource status error')
        return rh.reply_fail2(
            rh.REPLY_SERVICE_UNAVAILABLE_503,
            error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
            error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

    ui_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, entry_point.type, entry_point.id)
    ui_jnode = fs.load_json2(ui_jnode_filepath)

    return rh.reply_done(ui_jnode['rw'])


def _handle_put(ui_pid, entry_point, new_ui_jnode):
    """
    :type ui_pid: str
    :type entry_point: ph.UiIdentity
    :type new_ui_jnode: dict

    :rtype: dict
    """

    # check status of videosource
    videsource_status = helpers.get_videsource_status(ui_pid, entry_point.id)
    if videsource_status is None:
        eventlog.log_d('fail: get videosource status')
        return rh.reply_fail2(
            rh.REPLY_SERVICE_UNAVAILABLE_503,
            error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
            error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

    elif videsource_status != 0:
        eventlog.log_d('fail: videosource status error')
        return rh.reply_fail2(
            rh.REPLY_SERVICE_UNAVAILABLE_503,
            error_code=rh.FAIL_CODE_VIDEOSOURCE_ERROR,
            error_text=rh.FAIL_TEXT_VIDEOSOURCE_ERROR)

    restart_setupstore = []

    def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
        """
        :type ui_identity_to_jnode_map: dict of UiIdentity: dict
        :rtype: bool
        """
        if ui_identity_to_jnode_map[entry_point]['rw'] == new_ui_jnode:
            return False

        if ui_identity_to_jnode_map[entry_point]['rw']['enabled'] != new_ui_jnode['enabled']:
            restart_setupstore.append(True)

        ui_identity_to_jnode_map[entry_point]['rw'] = new_ui_jnode
        return True

    try:
        creator.generate(ui_pid, entry_point, _edit_ui_jnode_handler)

        if restart_setupstore:
            helpers.sighup_setupstore()
        return rh.reply_done(new_ui_jnode)

    except LockError:
        return rh.reply_fail(rh.REPLY_CONFLICT_409)


class UiHttpStreaming:
    def __init__(self, handler):
        """
        :type handler: object -> dict
        """

        self._handler = handler

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiHttpStreaming
        """
        #
        http_method = query_args.http_method
        url_query_string = query_args.query_string
        new_ui_jnode = ujson.loads(query_args.raw_body) if query_args.raw_body else None

        ui_pid = query_args.get_ui_pid()
        ui_type = query_args.get_ui_type()
        ui_id = query_args.get_ui_id()
        if None in [ui_pid, ui_type, ui_id]:
            return UiHttpStreaming(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiHttpStreaming(lambda: rh.reply_fail(rh.REPLY_NOT_FOUND_404))

        entry_point = ph.UiIdentity(ui_ptype, ui_type, ui_id)

        #
        if url_query_string:
            return UiHttpStreaming(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

        #
        if http_method == 'GET':
            if not new_ui_jnode:
                return UiHttpStreaming(lambda: _handle_get(ui_pid, entry_point))

        elif http_method == 'PUT':
            schema_jnode_filepath = ph.get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            request_schema_jnode = fs.load_json2(schema_jnode_filepath)
            if new_ui_jnode:
                try:
                    jsonschema.validate(new_ui_jnode, request_schema_jnode)
                    # validate "format" jnodes from json
                    footer_rows_jnode = new_ui_jnode['live']['footer']['rows']
                    if _validate_format(footer_rows_jnode) is False:
                        return UiHttpStreaming(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

                    return UiHttpStreaming(lambda: _handle_put(ui_pid, entry_point, new_ui_jnode))

                except (ValueError, jsonschema.ValidationError):
                    return UiHttpStreaming(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))
                except LockError:
                    return UiHttpStreaming(lambda: rh.reply_fail(rh.REPLY_CONFLICT_409))

        else:
            eventlog.log_d('fail: unsupported http-method')
            return UiHttpStreaming(lambda: rh.reply_fail(rh.REPLY_METHOD_NOT_ALLOWED_405))

        eventlog.log_d('fail: deserialize query')
        return UiHttpStreaming(lambda: rh.reply_fail(rh.REPLY_BAD_REQUEST_400))

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler()
