import jsonschema
import subprocess
import ujson
import hglib
import os

from edge import eventlog, path_helpers as ph
from edge.admin import reply_helpers as rh, ui_types as ui
from edge.com import fs

from ui_pipeline_set import get_pipeline_type
import edge_creator as creator


def _run_landmark(landmark_cmd):
    """
    :type landmark_cmd: str
    :rtype: dict
    """
    eventlog.log_d('landmark-cmd <%s>' % landmark_cmd)

    landmark = subprocess.Popen([landmark_cmd], stdout=subprocess.PIPE, shell=True)
    landmark_stdout = landmark.stdout.read()
    eventlog.log_d('landmark-out <%s>' % landmark_stdout)

    landmark_jnode = ujson.loads(landmark_stdout)
    return landmark_jnode


def _get_live_image(ui_pid):
    """
    :type ui_pid: str
    :rtype: str
    """
    # get last overlay imagepath
    or_store = "/var/spool/edge/pipeline-%s/or0-or0-store/" % ui_pid
    file_names = os.listdir(or_store)

    def unpack_ts(file_name):
        """
        :type file_name: str
        :rtype: int
        """
        return int(file_name.split('-')[-1])

    blob_name = max(file_names, key=unpack_ts)
    blob_path = os.path.join(or_store, blob_name)
    return blob_path


def _commit(ui_pid, jnode):
    """
    :type ui_pid: str
    :type jnode: dict
    :type: bool
    """

    client = hglib.open(ph.HG_DIR_PATH)
    jnode_path = ph.get_hg_jnode_file_path(ui_pid, ph.UI_TYPE_LOCATION)

    first_time = not os.path.exists(jnode_path)

    fs.write_json_file(jnode, jnode_path, pretty=True)
    image_path = ph.get_hg_image_file_path(ui_pid)
    or_image_path = _get_live_image(ui_pid)
    fs.silent_cp(or_image_path, image_path)

    if first_time:
        client.add([jnode_path, image_path])

    client.commit(message="edit(%s/location.0)" % ui_pid, user="admin")


class UiLocation:
    def __init__(self, handler, handler_args):
        """
        :type handler: object -> dict
        :type handler_args: object
        """
        self._handler = handler
        self._handler_args = handler_args

    @staticmethod
    def create(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLocation
        """
        if query_args.get_ui_action():
            return UiLocation._deserialize_location_define_rotation(query_args)

        else:
            return UiLocation._deserialize_location(query_args)

    @staticmethod
    def _deserialize_location(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLocation
        """
        if query_args.http_method == 'GET':
            return UiLocation._deserialize_get_location(query_args)

        elif query_args.http_method == 'PUT':
            return UiLocation._deserialize_put_location(query_args)

        else:
            eventlog.log_d("fail: method-not-allowed")
            return UiLocation(handler=UiLocation._reply_fail, handler_args=rh.REPLY_METHOD_NOT_ALLOWED_405)

    @staticmethod
    def _deserialize_get_location(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLocation
        """
        if query_args.query_string or query_args.raw_body:
            eventlog.log_d('fail: bad-request')
            return UiLocation(handler=UiLocation._reply_fail, handler_args=rh.REPLY_BAD_REQUEST_400)

        ui_pid = query_args.get_ui_pid()
        ui_type = query_args.get_ui_type()
        ui_id = query_args.get_ui_id()

        ui_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, ui_type, ui_id)
        return UiLocation(UiLocation._handle_get_location, ui_jnode_filepath)

    @staticmethod
    def _deserialize_put_location(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        :rtype: UiLocation
        """
        if query_args.query_string:
            eventlog.log_d('fail: bad-request')
            return UiLocation(handler=UiLocation._reply_fail, handler_args=rh.REPLY_BAD_REQUEST_400)

        # load new-ui-jnode
        new_ui_jnode = ujson.loads(query_args.raw_body) if query_args.raw_body else None
        if not new_ui_jnode:
            eventlog.log_d('fail: bad-request')
            return UiLocation(handler=UiLocation._reply_fail, handler_args=rh.REPLY_BAD_REQUEST_400)

        ui_pid = query_args.get_ui_pid()
        ui_type = query_args.get_ui_type()
        ui_id = query_args.get_ui_id()

        # load ui-schena-jnode
        try:
            ui_schema_filepath = ph.get_ui_pipeline_schema_filepath(ui_pid, ui_type, ui_id)
            with open(ui_schema_filepath) as json_file:
                ui_schema_jnode = ujson.load(json_file)

        except (IOError, ValueError) as x:
            eventlog.log_d("fail: load ui-schema-jnode (%s)" % x.message)
            return UiLocation(handler=UiLocation._reply_fail, handler_args=rh.REPLY_INTERNAL_SERVER_ERROR_500)

        # validate new-ui-jnode
        try:
            jsonschema.validate(new_ui_jnode, ui_schema_jnode)

        except (ValueError, jsonschema.ValidationError) as x:
            eventlog.log_d("fail: validate new-ui-jnode (%s)" % x.message)
            return UiLocation(handler=UiLocation._reply_fail, handler_args=rh.REPLY_BAD_REQUEST_400)

        # all arguments are valid
        return UiLocation(UiLocation._handle_put_location, (ui_pid, ui_type, ui_id, new_ui_jnode))

    @staticmethod
    def _deserialize_location_define_rotation(query_args):
        """
        :type query_args: edge.admin.ui_types.UiQueryArgs
        """
        if query_args.get_ui_action() != 'rotation':
            eventlog.log_d('fail: not-found')
            return UiLocation(handler=UiLocation._reply_fail, handler_args=rh.REPLY_NOT_FOUND_404)

        if query_args.raw_body or not query_args.query_string:
            eventlog.log_d('fail: bad-request')
            return UiLocation(handler=UiLocation._reply_fail, handler_args=rh.REPLY_BAD_REQUEST_400)

        if query_args.http_method != 'GET':
            eventlog.log_d("fail: method-not-allowed")
            return UiLocation(handler=UiLocation._reply_fail, handler_args=rh.REPLY_METHOD_NOT_ALLOWED_405)

        ui_pid = query_args.get_ui_pid()
        ui_id = query_args.get_ui_id()
        videosource_jnode_filepath = ph.get_ui_pipeline_jnode_filepath(ui_pid, ph.UI_TYPE_VIDEOSOURCE_MV, ui_id)
        videosource_jnode = fs.load_json2(videosource_jnode_filepath)
        fov_h, fov_v = videosource_jnode['ro']['fov-hv']

        rlandmark_cmd = UiLocation._mk_rlandmark_cmd(query_args.query_string, fov_h, fov_v)
        if not rlandmark_cmd:
            eventlog.log_d('fail: bad-request')
            return UiLocation(handler=UiLocation._reply_fail, handler_args=rh.REPLY_BAD_REQUEST_400)

        return UiLocation(UiLocation._handle_get_location_define_rotation, rlandmark_cmd)

    @staticmethod
    def _handle_get_location(ui_jnode_filepath):
        """
        :type ui_jnode_filepath: str
        """
        ui_jnode = fs.load_json2(ui_jnode_filepath)
        return rh.reply_done(ui_jnode['rw'], False)

    @staticmethod
    def _check_ui_jnode_changed(old_ui_jnode, new_ui_jnode):
        return old_ui_jnode != new_ui_jnode

    @staticmethod
    def _handle_put_location(handler_args):
        """
        :type handler_args: (str, str, str, dict)
        """
        ui_pid, ui_type, ui_id, new_ui_jnode = handler_args

        ui_ptype = get_pipeline_type(ui_pid)
        if ui_ptype is None:
            return UiLocation._reply_fail(rh.REPLY_INTERNAL_SERVER_ERROR_500)

        entry_point = ph.UiIdentity(ui_ptype, ui_type, ui_id)

        def _edit_ui_jnode_handler(ui_identity_to_jnode_map):
            """
            :type ui_identity_to_jnode_map: dict of UiIdentity: dict
            :rtype: bool
            """
            old_ui_jnode = ui_identity_to_jnode_map[entry_point]
            if old_ui_jnode['rw'] == new_ui_jnode:
                return False

            old_ui_jnode['rw'] = new_ui_jnode
            _commit(ui_pid, old_ui_jnode)

            custom_matrix = old_ui_jnode['ro'].get('rhcc-to-ndc')
            if custom_matrix is None:
                cam_ui_jnode = ui_identity_to_jnode_map[ph.UiIdentity(ui_ptype, ph.UI_TYPE_VIDEOSOURCE_MV, ui_id)]
                cam_fov_h, cam_fov_v = cam_ui_jnode['ro']['fov-hv']

                landmark_cmd = UiLocation._mk_landmark_cmd(new_ui_jnode, cam_fov_h, cam_fov_v)
                landmark_out = _run_landmark(landmark_cmd)
                old_ui_jnode['rt'] = {
                    'rhcc-to-ndc': landmark_out['rhcc-to-ndc']
                }

            return True

        try:
            creator.generate(ui_pid, entry_point, _edit_ui_jnode_handler)

            return rh.reply_done(new_ui_jnode, False)

        except ui.LockError:
            return UiLocation._reply_fail(rh.REPLY_CONFLICT_409)

    @staticmethod
    def _mk_landmark_cmd(new_ui_jnode, cam_fov_h, cam_fov_v):
        """
        :type new_ui_jnode: dict
        :type cam_fov_h: float
        :type cam_fov_v: float
        :rtype: str
        """

        camera_jnode = new_ui_jnode['args']['camera']
        # deserialize sensor-position
        position_jnode = camera_jnode['position']
        unit_f = position_jnode['forward']
        unit_l = position_jnode['left']
        unit_u = position_jnode['up']

        # deserialize sensor-rotation
        rotation_jnode = camera_jnode['rotation']
        unit_fl = rotation_jnode['azimuth']
        unit_fr = rotation_jnode['elevation']
        unit_lu = rotation_jnode.get('roll', 0)

        name_to_value_map = {
            '--fov-vertical': cam_fov_v,
            '--fov-horizontal': cam_fov_h,

            '--position-forward': unit_f,
            '--position-left': unit_l,
            '--position-top': unit_u,

            '--azimuth': unit_fl,
            '--elevation': unit_fr,
            '--roll': unit_lu
        }

        command_line = '/opt/edge/bin/landmark_v2 --mode-landmark'
        for name, value in name_to_value_map.iteritems():
            command_line += ' %s=%s' % (name, value)

        return command_line

    @staticmethod
    def _mk_rlandmark_cmd(query_string, cam_fov_h, cam_fov_v):
        """
        :type query_string: str
        :rtype: str | None
        """
        if not (query_string and query_string.startswith('?')):
            eventlog.log_d('fail: invalid arg')
            return None

        args = {
            '--fov-vertical': cam_fov_v,
            '--fov-horizontal': cam_fov_h,

            '--position-forward': None,
            '--position-left': None,
            '--position-top': None,

            '--rotation-landmark-image': None,
            '--rotation-landmark-forward': None,
            '--rotation-landmark-left': None
        }

        for name_value in query_string[1:].split('&'):
            name, value = name_value.split('=', 1)

            if name == 'camera-forward':
                args['--position-forward'] = float(value)

            elif name == 'camera-left':
                args['--position-left'] = float(value)

            elif name == 'camera-up':
                args['--position-top'] = float(value)

            elif name == 'landmark-image':
                if value in ['bottom-left', 'bottom-right']:
                    args['--rotation-landmark-image'] = value

            elif name == 'landmark-forward':
                args['--rotation-landmark-forward'] = float(value)

            elif name == 'landmark-left':
                args['--rotation-landmark-left'] = float(value)

        invalid_args = filter(lambda (arg_name, arg_value): arg_value is None,
                              args.iteritems())
        if invalid_args:
            for invalid_arg_name, _ in invalid_args:
                eventlog.log_d('fail: param "%s" is absent' % invalid_arg_name)
            return None

        else:
            # tweak landmark coordinates relative to camera position
            args['--rotation-landmark-forward'] -= args['--position-forward']
            args['--rotation-landmark-left'] -= args['--position-left']

            command_line = '/opt/edge/bin/landmark_v2 --mode-rlandmark'
            for name, value in args.iteritems():
                command_line += ' %s=%s' % (name, value)
            return command_line

    @staticmethod
    def _handle_get_location_define_rotation(landmark_cmd):
        """
        :type landmark_cmd: str
        :rtype: dict
        """
        landmark_jnode = _run_landmark(landmark_cmd)
        calculated_rot = landmark_jnode['unit-rotation-auto']
        jnode = {
            "camera": {
                "rotation": {
                    "elevation": calculated_rot['elevation'],
                    "azimuth": calculated_rot['azimuth']
                }
            }
        }
        return rh.reply_done(jnode, False)

    @staticmethod
    def _reply_fail(err_code):
        """
        :type err_code: int
        :rtype: dict
        """
        return rh.reply_fail(err_code, False)

    def handle(self):
        """
        :rtype: dict
        """
        return self._handler(self._handler_args)
