import os
import subprocess

import edge.com.fs as fs
import edge.path_helpers as ph
from edge import eventlog
from tempfile import NamedTemporaryFile


ENV = os.environ.copy()
upd = {'OPENSSL_CONF': '/etc/edge/openssl1.0/openssl.cnf'}
ENV.update(upd)


def run_silent(esc_cmd, known_errors=list()):
    """
    :param esc_cmd: Escaped shell-command. Use pipes.quote() to escape input form external sources
    :type esc_cmd: str,
    :type known_errors: list of str
    :rtype: (int, str)| None
    """

    try:
        # eventlog.log_d("exec <%s>" % esc_cmd)
        p = subprocess.Popen([esc_cmd], env=ENV, stderr=subprocess.PIPE, stdout=subprocess.PIPE, shell=True)
        out, err = p.communicate()
        rc = p.returncode

        # eventlog.log_d("<%s> <%s>" % (out, err))
        if len(err) == 0 or err in known_errors:
            return rc, out
        else:
            return None

    except Exception as x:
        eventlog.log_e('fail: <%s>' % x.message)

    return None


def check(crt_filepath, key_filepath, key_password):
    """
    :type crt_filepath: str
    :type key_filepath: str
    :type key_password: str
    :rtype: bool
    """

    valid = False

    for keytype in ['pk8', 'dat', 'pfx', 'pem']:
        if valid:
            break

        key_public_key = 'random1'

        if keytype == 'pk8':
            tmp_key = NamedTemporaryFile()
            reply = run_silent('openssl1.0 pkcs8 -in %s -inform DER -out %s -passin pass:%s' %
                               (key_filepath, tmp_key.name, key_password))
            if reply is None:
                continue

            rc, out = reply
            if rc < 0:
                continue

            reply = run_silent('openssl1.0 pkey -pubout -in %s' % tmp_key.name)
            if reply is None:
                continue

            rc, key_public_key = reply
            if rc < 0:
                continue

        elif keytype == 'dat':
            continue

        elif keytype == 'pfx':
            tmp_key = NamedTemporaryFile()
            reply = run_silent('openssl1.0 pkcs12 -in %s -nocerts -out %s -passin pass:%s -nodes' %
                               (key_filepath, tmp_key.name, key_password),
                               ['MAC verified OK\n'])
            if reply is None:
                continue

            rc, out = reply
            if rc < 0:
                continue

            reply = run_silent('openssl1.0 pkey -pubout -in %s' % tmp_key.name)
            if reply is None:
                continue

            rc, key_public_key = reply
            if rc < 0:
                continue

        elif keytype == 'pem':
            reply = run_silent('openssl1.0 pkey -pubout -in %s' % key_filepath)
            if reply is None:
                continue

            rc, key_public_key = reply
            if rc < 0:
                continue

        for crttype in ['crt', 'cer']:
            crt_public_key = 'random2'

            if crttype == 'crt':
                reply = run_silent('openssl1.0 x509 -pubkey -in %s -noout' % crt_filepath)
                if reply is None:
                    continue

                rc, crt_public_key = reply
                if rc < 0:
                    continue

            elif crttype == 'cer':
                tmp_crt = NamedTemporaryFile()
                reply = run_silent('openssl1.0 x509 -inform DER -in %s -out %s' % (crt_filepath, tmp_crt.name))
                if reply is None:
                    continue

                rc, stdout = reply
                if rc < 0:
                    continue

                reply = run_silent('openssl1.0 x509 -pubkey -in %s -noout' % tmp_crt.name)

                if reply is None:
                    continue

                rc, crt_public_key = reply
                if rc < 0:
                    continue

            if crt_public_key == key_public_key:
                eventlog.log_d("valid for key <%s> certificate <%s>" % (keytype, crttype))
                valid = True
                break

            else:
                eventlog.log_e("fail: private key and certificate have different public keys")
                # eventlog.log_d("key <%s> certificate <%s>" % (key_public_key, crt_public_key))

    return valid


class Result:
    def __init__(self, key_file, crt_file, key_pass):
        """
        :type key_file: file
        :type crt_file: file
        :type key_pass: str
        """
        self.key = key_file
        self.crt = crt_file
        self.key_pass = key_pass


def check2(crt_filepath, key_filepath, key_password):
    """
    :type crt_filepath: str
    :type key_filepath: str
    :type key_password: str
    :rtype: Result | None
    """

    pem_key_file = NamedTemporaryFile(prefix='tmpkey', suffix='.key', dir='/tmp/edge')
    cer_certificate_file = NamedTemporaryFile(prefix='tmpcrt', suffix='.crt', dir='/tmp/edge')

    for keytype in ['pk8', 'dat', 'pfx', 'pem']:

        key_public_key = 'random1'

        if keytype == 'pk8':
            tmp_key = pem_key_file.name
            reply = run_silent('openssl1.0 pkcs8 -in %s -inform DER -out %s -passin pass:%s' %
                               (key_filepath, tmp_key, key_password))
            if reply is None:
                continue

            rc, out = reply
            if rc < 0:
                continue

            reply = run_silent('openssl1.0 pkey -pubout -in %s' % tmp_key)
            if reply is None:
                continue

            rc, key_public_key = reply
            if rc < 0:
                continue

        elif keytype == 'dat':
            continue

        elif keytype == 'pfx':
            tmp_key = pem_key_file.name
            reply = run_silent('openssl1.0 pkcs12 -in %s -nocerts -out %s -passin pass:%s -nodes' %
                               (key_filepath, tmp_key, key_password),
                               ['MAC verified OK\n'])
            if reply is None:
                continue

            rc, out = reply
            if rc < 0:
                continue

            reply = run_silent('openssl1.0 pkey -pubout -in %s' % tmp_key)
            if reply is None:
                continue

            rc, key_public_key = reply
            if rc < 0:
                continue

            fs.silent_cp(key_filepath, pem_key_file.name)

        elif keytype == 'pem':
            reply = run_silent('openssl1.0 pkey -pubout -in %s' % key_filepath)
            if reply is None:
                continue

            rc, key_public_key = reply
            if rc < 0:
                continue

            fs.silent_cp(key_filepath, pem_key_file.name)

        for crttype in ['crt', 'cer']:
            crt_public_key = 'random2'

            if crttype == 'crt':
                reply = run_silent('openssl1.0 x509 -pubkey -in %s -noout' % crt_filepath)
                if reply is None:
                    continue

                rc, crt_public_key = reply
                if rc < 0:
                    continue

                fs.silent_cp(crt_filepath, cer_certificate_file.name)

            elif crttype == 'cer':
                tmp_crt = cer_certificate_file.name
                reply = run_silent('openssl1.0 x509 -inform DER -in %s -out %s' % (crt_filepath, tmp_crt))
                if reply is None:
                    continue

                rc, stdout = reply
                if rc < 0:
                    continue

                reply = run_silent('openssl1.0 x509 -pubkey -in %s -noout' % tmp_crt)

                if reply is None:
                    continue

                rc, crt_public_key = reply
                if rc < 0:
                    continue

            if crt_public_key == key_public_key:
                eventlog.log_d("valid for key <%s> certificate <%s>" % (keytype, crttype))
                return Result(pem_key_file, cer_certificate_file, key_password)

            else:
                eventlog.log_e("fail: private key and certificate have different public keys")
                eventlog.log_d("key <%s> certificate <%s>" % (key_public_key, crt_public_key))

    return None


def _get_limits_jnode(ui_jnode):
    """
    :type ui_jnode: dict
    :rtype: dict
    """
    ro_jnode = ui_jnode.get('ro')

    crt_file = ro_jnode.get('crt') or False
    key_file = ro_jnode.get('key') or False

    if crt_file and key_file:
        return {'activatable': True}

    return {'activatable': False}


def get():
    """
    :rtype: dict | None
    """
    if not os.path.exists(ph.SIGN_CONF_FILEPATH):
        fs.create_jnode_from_default(ph.SIGN_DEFAULT_CONF_FILEPATH, ph.SIGN_CONF_FILEPATH)
    sign_jnode = fs.load_json(ph.SIGN_CONF_FILEPATH)
    if sign_jnode:
        reply_jnode = sign_jnode.get('rw')
        reply_jnode['limits'] = _get_limits_jnode(sign_jnode)
        return reply_jnode

    return None


def put(body):
    """
    :type body: dict
    :rtype: dict | None
    """
    sign_jnode = fs.load_json(ph.SIGN_CONF_FILEPATH)
    if sign_jnode is None:
        eventlog.log_e('fail: open ')
        return None

    reply_jnode = sign_jnode.get('rw')
    reply_jnode['args']['active'] = body['args']['active']

    fs.write_json_file(sign_jnode, ph.SIGN_CONF_FILEPATH, True)

    reply_jnode['limits'] = _get_limits_jnode(sign_jnode)
    return reply_jnode


def put_crt_key(crt_filepath, key_filepath, key_password):
    """
    :type crt_filepath: str
    :type key_filepath: str
    :type key_password: str | None
    :rtype: dict | 500 | 400
    """
    eventlog.log_d('try put crt %s key %s pass %s' % (crt_filepath, key_filepath, key_password))
    # for check
    if not check(crt_filepath, key_filepath, key_password):
        eventlog.log_e("fail: key and cert match")
        return 400

    # for check2
    # reply = check2(crt_filepath, key_filepath, key_password)
    # if reply is None:
    #     eventlog.log_e("fail: key and cert match")
    #     return 400

    sign_jnode = fs.load_json(ph.SIGN_CONF_FILEPATH)
    if sign_jnode is None:
        eventlog.log_e('fail: open %s' % sign_jnode)
        return 500

    # for check
    fs.silent_cp(crt_filepath, ph.SIGN_CRT_FILEPATH)
    fs.silent_cp(key_filepath, ph.SIGN_KEY_FILEPATH)

    # for check2
    # fs.silent_cp(reply.crt.name, ph.SIGN_CRT_FILEPATH)
    # fs.silent_cp(reply.key.name, ph.SIGN_KEY_FILEPATH)

    rw_jnode = sign_jnode.get('rw')
    ro_jnode = sign_jnode.get('ro')

    ro_jnode['key'] = True
    ro_jnode['crt'] = True
    if key_password:
        ro_jnode['key-sign-pass'] = key_password
    else:
        ro_jnode['key-sign-pass'] = ''

    fs.write_json_file(sign_jnode, ph.SIGN_CONF_FILEPATH, True)

    rw_jnode['limits'] = _get_limits_jnode(sign_jnode)
    return rw_jnode
