#!/usr/bin/python
import os
import shutil
import subprocess
import time
import ujson
import sys

import edge.admin.ui_types as ui_types
from edge import eventlog
import jsonschema
from edge.admin.reply_helpers import *
from edge.com.fs import ln_schema, create_jnode_from_default, try_lock_pid_file

HTTP_GET = 'GET'
HTTP_PUT = 'PUT'
HTTP_POST = 'POST'
HTTP_DELETE = 'DELETE'

UI_USERS_SCHEMA_FILEPATH = '/etc/edge/setupstore/schemas/user-password.json'
UI_USERS_JNODE_FILEPATH = '/etc/edge/setupstore/admin/actions/user.json'
UI_USERS_DEFJNODE_FILEPATH = '/opt/edge/share/setupstore/ui/user-password-jnode.json'

LOCK_FILEPATH = '/var/run/edge/user-password'

PASSWD_CONF_PATH = '/etc/edge/webserver.0/htpasswd'

RESTART_NONE = 0
RESTART_WEBSERVER = 1

BACKUP_DIRPATH = '/etc/edge/setupstore/backup/user-password.' + \
                 str(round(time.time() * 1000000)) + '/'


def try_create_backup():
    # noinspection PyBroadException
    try:
        os.makedirs(BACKUP_DIRPATH)
        shutil.copy(PASSWD_CONF_PATH, BACKUP_DIRPATH)
        return True

    except Exception:
        pass

    return False


def try_restore_backup():
    # noinspection PyBroadException
    try:
        shutil.copy(BACKUP_DIRPATH + 'htpasswd', PASSWD_CONF_PATH)
        os.rmdir(BACKUP_DIRPATH)
        return True

    except Exception:
        return False


def remove_backup():
    # noinspection PyBroadException
    try:
        shutil.rmtree(BACKUP_DIRPATH)
    except Exception:
        pass


class BadRequestException(RuntimeError):
    pass


def edit_passwd_conf(user, new_passw):
    _PROHIBITED_SYMBOLS = ('\\', "'")
    if any(sym in new_passw for sym in _PROHIBITED_SYMBOLS):
        eventlog.log_e('fail: prohibited symbols in password')
        raise BadRequestException

    with open(PASSWD_CONF_PATH, 'r') as old_ntp_file:
        old_passwd_lines = old_ntp_file.readlines()

    new_passwd_lines = []
    if len(old_passwd_lines) > 0:
        for old_line in old_passwd_lines:
            if old_line.startswith(user):
                continue

            new_passwd_lines.append(old_line)

        new_passw = new_passw.replace('"', '\\"')
        password = subprocess.Popen('openssl passwd -apr1 "{}"'.format(new_passw),
                                    shell=True, stdout=subprocess.PIPE)
        new_password = password.communicate()[0]
        new_passwd_lines.append('%s:%s' % (user, new_password))

    else:
        raise NotImplemented

    with open(PASSWD_CONF_PATH, 'w') as new_passwd_file:
        new_passwd_file.write(''.join(new_passwd_lines))


def edit_passwd(user, new_passwd):
    edit_passwd_conf(user, new_passwd)
    return RESTART_WEBSERVER


def restart_module(module):
    subprocess.Popen(["kill -9 `ps ax |"
                      "grep \"%s --id.0\" |"
                      "grep -v grep |"
                      "awk '{print $1}'`" % module],
                     shell=True)


def _handle_get(query_args):
    """
    :type query_args: UiQueryArgs
    :rtype: dict
    """
    if query_args.path_info:
        eventlog.log_d('fail: not-found')
        return reply_fail(REPLY_NOT_FOUND_404)

    if query_args.query_string:
        eventlog.log_d('fail: bad-request')
        return reply_fail(REPLY_BAD_REQUEST_400)

    # create ui-jnode
    if not os.path.exists(UI_USERS_JNODE_FILEPATH):
        # noinspection PyBroadException
        try:
            create_jnode_from_default(UI_USERS_DEFJNODE_FILEPATH, UI_USERS_JNODE_FILEPATH)

        except Exception:
            eventlog.log_x()
            return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    # create ui-schema
    if not os.path.islink(UI_USERS_SCHEMA_FILEPATH) or not os.path.exists(os.readlink(UI_USERS_SCHEMA_FILEPATH)):
        if not os.path.exists(UI_USERS_JNODE_FILEPATH):
            eventlog.log_d('fail: %s not found' % UI_USERS_JNODE_FILEPATH)
            return reply_fail(REPLY_NOT_FOUND_404)

        ln_schema(UI_USERS_SCHEMA_FILEPATH, UI_USERS_JNODE_FILEPATH)

    try:
        with open(UI_USERS_JNODE_FILEPATH, 'r') as json_file:
            ui_jnode = ujson.load(json_file)
        return reply_done(ui_jnode['rw']['users'])

    except (IOError, ValueError) as x:
        eventlog.log_d('fail: load ui-jnode (%s)' % x.message)
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)


def _handle_put(user, new_password):
    if not try_lock_pid_file(LOCK_FILEPATH):
        return reply_fail(REPLY_CONFLICT_409)

    if not try_create_backup():
        return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

    success = False
    # noinspection PyBroadException
    try:
        restart_mask = 0
        restart_mask |= edit_passwd(user, new_password)

        if restart_mask & RESTART_WEBSERVER:
            restart_module('webserver')

        success = True

    except BadRequestException:
        eventlog.log_d('lala')
        if try_restore_backup():
            remove_backup()
            eventlog.log_d('lalala')
        return reply_fail(REPLY_BAD_REQUEST_400)

    except Exception:
        pass

    if success:
        remove_backup()
        return reply_done([])

    if try_restore_backup():
        remove_backup()

    response_jnode = reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)
    return response_jnode


def _run(query_args):
    """
    :param query_args: UiQueryArgs
    :return:
    """
    # Dispatch http-method.
    if query_args.http_method == 'GET':
        return _handle_get(query_args)

    elif query_args.http_method == 'PUT':
        new_jnode = ujson.loads(query_args.raw_body)
        user = query_args.get_path_info_item(0)
        if not new_jnode or not user:
            eventlog.log_d('fail: not enough request params')
            return reply_fail(REPLY_BAD_REQUEST_400)

        try:
            with open(UI_USERS_SCHEMA_FILEPATH, 'r') as json_file:
                request_schema_jnode = ujson.load(json_file)
        except (IOError, ValueError):
            eventlog.log_d('fail: load ui-schema')
            return reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)

        try:
            jsonschema.validate(new_jnode, request_schema_jnode)
        except jsonschema.ValidationError:
            eventlog.log_x()
            return reply_fail(REPLY_BAD_REQUEST_400)

        except ValueError as x:
            eventlog.log_d('fail: (%s)' % x.message)
            return reply_fail(REPLY_BAD_REQUEST_400)

        response_jnode = _handle_put(user, new_jnode['password'])

    else:
        response_jnode = reply_fail(REPLY_METHOD_NOT_ALLOWED_405)

    return response_jnode


# noinspection PyBroadException
def _main():
    eventlog.open_log("cgi-users", "debug")
    exit_code = 1
    try:
        query_args = ui_types.UiQueryArgs.deserialize()
        reply_jnode = _run(query_args)

        print ujson.dumps(reply_jnode)
        exit_code = 0

    except Exception:
        eventlog.log_x()
        print ujson.dumps(reply_fail(REPLY_INTERNAL_SERVER_ERROR_500))

    finally:
        eventlog.close_log()

    sys.exit(exit_code)


if __name__ == '__main__':
    _main()
